//////////////////////////////////////////////////////////////////////////////
//   Author:       George Potts
//   Copyright:    PottsSoftware
//   Initial Date: 2015/08/17
// 
//   Description:  Split into separate file and fleshed out Add/Sub/Adds/Subs
//
//   Notes:
//
//   Revision History:
//
//////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "SSELibrary.h"
#include <immintrin.h>

#pragma region Adds_U8
// SSE2
SSELIBRARY_API void SSE_Adds_U8(char* dst, char* src, size_t bytes)
{
	size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
	__m128i * ptrDst = (__m128i*)dst;
	__m128i * ptrSrc = (__m128i*)src;

    for (; loops > 0; loops--)
	{
		__m128i m_dst = _mm_load_si128(ptrDst);
		__m128i m_src = _mm_load_si128(ptrSrc);
		__m128i result = _mm_adds_epu8(m_dst, m_src);
		_mm_store_si128(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}

// AVX2
SSELIBRARY_API void AVX2_Adds_U8(char* dst, char* src, size_t bytes)
{
	size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
	__m256i * ptrDst = (__m256i*)dst;
	__m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
	{
		__m256i m_dst = _mm256_loadu_si256(ptrDst);
		__m256i m_src = _mm256_loadu_si256(ptrSrc);
		__m256i result = _mm256_adds_epu8(m_dst, m_src);
		_mm256_store_si256(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}

SSELIBRARY_API void CPP_Adds_U8(char* dst, char* src, size_t bytes)
{
	unsigned int val;
	unsigned char * pDst = (unsigned char*)dst;
	unsigned char * pSrc = (unsigned char*)src;

    for (; bytes > 0; bytes--)
	{
		val = (unsigned int)*pSrc + *pDst;
		if (val & 0x100) *pDst = 255;
		else *pDst = (unsigned char)val;
        pSrc++;
        pDst++;
	}
}
#pragma endregion Adds_U8

#pragma region Adds_I8
// SSE2
SSELIBRARY_API void SSE_Adds_I8(char* dst, char* src, size_t bytes)
{
	size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
	__m128i * ptrDst = (__m128i*)dst;
	__m128i * ptrSrc = (__m128i*)src;

    for (; loops > 0; loops--)
	{
		__m128i m_dst = _mm_load_si128(ptrDst);
		__m128i m_src = _mm_load_si128(ptrSrc);
		__m128i result = _mm_adds_epi8(m_dst, m_src);
		_mm_store_si128(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}

// AVX2
SSELIBRARY_API void AVX2_Adds_I8(char* dst, char* src, size_t bytes)
{
	size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
	__m256i * ptrDst = (__m256i*)dst;
	__m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
	{
		__m256i m_dst = _mm256_loadu_si256(ptrDst);
		__m256i m_src = _mm256_loadu_si256(ptrSrc);
		__m256i result = _mm256_adds_epi8(m_dst, m_src);
		_mm256_store_si256(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}

// I *believe* that we're contrained to min and max vals on signed saturation math.
// TODO: confirm vs the HW implementation.
SSELIBRARY_API void CPP_Adds_I8(char* dst, char* src, size_t bytes)
{
	signed int val;
	signed char * pDst = (signed char*)dst;
	signed char * pSrc = (signed char*)src;
    for (; bytes > 0; bytes--)
	{
		val = (signed int)*pSrc + *pDst;
		if (val >= 0x80) *pDst = 127;
		else if (val < -128) *pDst = -128;
		else *pDst = (signed char)val;
        pSrc++;
        pDst++;
	}
}
#pragma endregion Adds_I8

#pragma region Subs_U8
// SSE2
SSELIBRARY_API void SSE_Subs_U8(char* dst, char* src, size_t bytes)
{
	size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
	__m128i * ptrDst = (__m128i*)dst;
	__m128i * ptrSrc = (__m128i*)src;

    for (; loops > 0; loops--)
	{
		__m128i m_dst = _mm_load_si128(ptrDst);
		__m128i m_src = _mm_load_si128(ptrSrc);
		__m128i result = _mm_subs_epu8(m_dst, m_src);
		_mm_store_si128(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}

// AVX2
SSELIBRARY_API void AVX2_Subs_U8(char* dst, char* src, size_t bytes)
{
	size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
	__m256i * ptrDst = (__m256i*)dst;
	__m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
	{
		__m256i m_dst = _mm256_loadu_si256(ptrDst);
		__m256i m_src = _mm256_loadu_si256(ptrSrc);
		__m256i result = _mm256_subs_epu8(m_dst, m_src);
		_mm256_store_si256(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}

SSELIBRARY_API void CPP_Subs_U8(char* dst, char* src, size_t bytes)
{
	signed int val;
	unsigned char * pDst = (unsigned char*)dst;
	unsigned char * pSrc = (unsigned char*)src;
    for (; bytes > 0; bytes--)
	{
		val = (int)*pDst - *pSrc;
		if (val < 0)*pDst = 0;
		else *pDst = (unsigned char)val;
        pSrc++;
        pDst++;
	}
}
#pragma endregion Subs_U8

#pragma region Subs_I8
// SSE2
SSELIBRARY_API void SSE_Subs_I8(char* dst, char* src, size_t bytes)
{
	size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
	__m128i * ptrDst = (__m128i*)dst;
	__m128i * ptrSrc = (__m128i*)src;

    for (; loops > 0; loops--)
	{
		__m128i m_dst = _mm_load_si128(ptrDst);
		__m128i m_src = _mm_load_si128(ptrSrc);
		__m128i result = _mm_subs_epi8(m_dst, m_src);
		_mm_store_si128(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}

// AVX2
SSELIBRARY_API void AVX2_Subs_I8(char* dst, char* src, size_t bytes)
{
	size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
	__m256i * ptrDst = (__m256i*)dst;
	__m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
	{
		__m256i m_dst = _mm256_loadu_si256(ptrDst);
		__m256i m_src = _mm256_loadu_si256(ptrSrc);
		__m256i result = _mm256_subs_epi8(m_dst, m_src);
		_mm256_store_si256(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}

// I *believe* that we're contrained to min and max vals on signed saturation math.
// TODO: confirm vs the HW implementation.
SSELIBRARY_API void CPP_Subs_I8(char* dst, char* src, size_t bytes)
{
	signed int val;
	signed char * pDst = (signed char*)dst;
	signed char * pSrc = (signed char*)src;

    for (; bytes > 0; bytes--)
	{
		val = (signed int)*pDst - *pSrc;
		if (val < -128) *pDst = -128;
		else if (val > 127) *pDst = 127;
		else *pDst = (signed char)val;

        pSrc++;
        pDst++;
	}
}
#pragma endregion Subs_I8

#pragma region Sub_I8
// SSE2
SSELIBRARY_API void SSE_Sub_I8(char* dst, char* src, size_t bytes)
{
	size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
	__m128i * ptrDst = (__m128i*)dst;
	__m128i * ptrSrc = (__m128i*)src;

    for (; loops > 0; loops--)
	{
		__m128i m_dst = _mm_load_si128(ptrDst);
		__m128i m_src = _mm_load_si128(ptrSrc);
		__m128i result = _mm_sub_epi8(m_dst, m_src);
		_mm_store_si128(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}

// AVX2
SSELIBRARY_API void AVX2_Sub_I8(char* dst, char* src, size_t bytes)
{
	size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
	__m256i * ptrDst = (__m256i*)dst;
	__m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
	{
		__m256i m_dst = _mm256_load_si256(ptrDst);
		__m256i m_src = _mm256_load_si256(ptrSrc);
		__m256i result = _mm256_sub_epi8(m_dst, m_src);
		_mm256_store_si256(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}

SSELIBRARY_API void CPP_Sub_I8(char* dst, char* src, size_t bytes)
{
	signed char * pDst = (signed char*)dst;
	signed char * pSrc = (signed char*)src;

    for (; bytes > 0; bytes--)
	{
		*pDst = *pDst - *pSrc;
        pSrc++;
        pDst++;
	}
}
#pragma endregion Sub_I8

#pragma region Sub_U8
SSELIBRARY_API void SSE_Sub_U8(char* dst, char* src, size_t bytes)
{
	SSE_Sub_I8(dst, src, bytes);
}

SSELIBRARY_API void AVX2_Sub_U8(char* dst, char* src, size_t bytes)
{
	AVX2_Sub_I8(dst, src, bytes);
}

SSELIBRARY_API void CPP_Sub_U8(char* dst, char* src, size_t bytes)
{
	CPP_Sub_I8(dst, src, bytes);
}
#pragma endregion Sub_U8

#pragma region Add_I8
// SSE2
SSELIBRARY_API void SSE_Add_I8(char* dst, char* src, size_t bytes)
{
	size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
	__m128i * ptrDst = (__m128i*)dst;
	__m128i * ptrSrc = (__m128i*)src;

    for (; loops > 0; loops--)
	{
		__m128i m_dst = _mm_load_si128(ptrDst);
		__m128i m_src = _mm_load_si128(ptrSrc);
		__m128i result = _mm_add_epi8(m_dst, m_src);
		_mm_store_si128(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}

// AVX2
SSELIBRARY_API void AVX2_Add_I8(char* dst, char* src, size_t bytes)
{
	size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
	__m256i * ptrDst = (__m256i*)dst;
	__m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
	{
		__m256i m_dst = _mm256_load_si256(ptrDst);
		__m256i m_src = _mm256_load_si256(ptrSrc);
		__m256i result = _mm256_add_epi8(m_dst, m_src);
		_mm256_store_si256(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}

SSELIBRARY_API void CPP_Add_I8(char* dst, char* src, size_t bytes)
{
	signed char * pDst = (signed char*)dst;
	signed char * pSrc = (signed char*)src;
    for (; bytes > 0; bytes--)
	{
		*pDst = *pDst + *pSrc;
        pSrc++;
        pDst++;
	}
}
#pragma endregion Add_I8

#pragma region Add_U8
SSELIBRARY_API void SSE_Add_U8(char* dst, char* src, size_t bytes)
{
	SSE_Add_I8(dst, src, bytes);
}

SSELIBRARY_API void AVX2_Add_U8(char* dst, char* src, size_t bytes)
{
	AVX2_Add_I8(dst, src, bytes);
}

SSELIBRARY_API void CPP_Add_U8(char* dst, char* src, size_t bytes)
{
	CPP_Add_I8(dst, src, bytes);
}
#pragma endregion Add_U8