//////////////////////////////////////////////////////////////////////////////
//   Author:       George Potts
//   Copyright:    PottsSoftware
//   Initial Date: 2015/08/13
// 
//   Description:  Implements many of the operations for 8-bit variables.
//
//   Notes:
//
//   Revision History:
//
//////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "SSELibrary.h"
#include <immintrin.h>

SSELIBRARY_API void SSE_Avg_U8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    for( ; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);
        __m128i result = _mm_avg_epu8(m_dst, m_src);
        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

// Requires AVX2
SSELIBRARY_API void AVX2_Avg_U8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for( ; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);
        __m256i result = _mm256_avg_epu8(m_dst, m_src);
        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void CPP_Avg_U8(char* dst, char* src, size_t bytes)
{
	unsigned char * pdst = (unsigned char*)dst;
	unsigned char * psrc = (unsigned char*)src;

    unsigned int val = 0;

    for( ; bytes > 0; bytes--)
	{
        val = (unsigned int)(*pdst + *psrc + 1);
        *pdst = (unsigned char)(val >> 1);
        psrc++;
        pdst++;
	}
}
