//////////////////////////////////////////////////////////////////////////////
//   Author:       George Potts
//   Copyright:    PottsSoftware
//   Initial Date: 2015/08/13
// 
//   Description:  Detect CPU features.
//
//   Notes:
//
//   Revision History:
//
//////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <immintrin.h>
#include <intrin.h>
#include "SSELibrary.h"
#include "CPUDetection.h"

DWORD GetMaxCall(DWORD mask);

CPUID_Info g_info;

// SEH
int filter(unsigned int code, struct _EXCEPTION_POINTERS * ep)
{
    UNREFERENCED_PARAMETER(code);
    UNREFERENCED_PARAMETER(ep);
    return EXCEPTION_CONTINUE_EXECUTION;
}

void GetCPUInfo(int function, int subFunction)
{
    memset(&g_info, 0, sizeof(g_info));
    __try
    {
        __cpuidex((int*)&g_info, function, subFunction);
    }
    __except(filter(GetExceptionCode(),GetExceptionInformation())){}
}

DWORD GetMaxCall(DWORD mask)
{
    GetCPUInfo((int)mask,0);
    return g_info.named.EAX;
}

SSELIBRARY_API bool Supports_MMX()
{
    GetCPUInfo(0x80000001,0);
    return g_info.regF8.edx.Intel.MMX == 1;
}

SSELIBRARY_API bool Supports_MMXEXT()
{
    GetCPUInfo(0x80000001,0);
    return g_info.regF8.edx.Intel.MMXEXT == 1;
}

SSELIBRARY_API bool Supports_3DNOW()
{
    GetCPUInfo(0x80000001,0);
    return g_info.regF8.edx.Intel._3DNOW == 1;
}

SSELIBRARY_API bool Supports_3DNOWEXT()
{
    GetCPUInfo(0x80000001,0);
    return g_info.regF8.edx.Intel._3DNOWEXT== 1;
}

SSELIBRARY_API bool Supports_SSE1()
{
    GetCPUInfo(1,0);
    return g_info.regF1.edx.Intel.SSE == 1;
}
SSELIBRARY_API bool Supports_SSE2()
{
    GetCPUInfo(1,0);
    return g_info.regF1.edx.Intel.SSE2 == 1;
}

SSELIBRARY_API bool Supports_SSE3()
{
    GetCPUInfo(1,0);
    return g_info.regF1.ecx.Intel.SSE3 == 1;
}

SSELIBRARY_API bool Supports_SSSE3()
{
    GetCPUInfo(1,0);
    return g_info.regF1.ecx.Intel.SSSE3 == 1;
}

SSELIBRARY_API bool Supports_SSE4A()
{
    GetCPUInfo(0x80000001,0);
    return g_info.regF8.ecx.Intel.SSE4A == 1;
}

SSELIBRARY_API bool Supports_SSE41()
{
    GetCPUInfo(1,0);
    return g_info.regF1.ecx.Intel.SSE41 == 1;
}

SSELIBRARY_API bool Supports_SSE42()
{
    GetCPUInfo(1,0);
    return g_info.regF1.ecx.Intel.SSE42 == 1;
}

SSELIBRARY_API bool Supports_AVX1()
{
    bool ret = false;

    // AVX needs to be supported by the CPU and the OS - need to test both.
    GetCPUInfo(1,0);
    bool avxFlag = g_info.regF1.ecx.Intel.AVX == 1;
    bool osxFlag = g_info.regF1.ecx.Intel.OSXSAVE == 1;
    if ( avxFlag && osxFlag )
    {
        // Supported by the CPU, check the OS.
        unsigned long long featureMask = _xgetbv(_XCR_XFEATURE_ENABLED_MASK);
        ret = (featureMask & 0x6) == 0x6;
    }
    return ret;
}

//
// https://software.intel.com/sites/default/files/article/405250/how-to-detect-new-instruction-support-in-the-4th-generation-intel-core-processor-family.pdf
// CPUID.(EAX = 01H, ECX = 0H) : ECX.FMA[bit 12]==1 &&
// CPUID.(EAX = 07H, ECX = 0H) : EBX.AVX2[bit 5] == 1 &&
// CPUID.(EAX = 07H, ECX = 0H) : EBX.BMI1[bit 3] == 1 &&
// CPUID.(EAX = 07H, ECX = 0H) : EBX.BMI2[bit 8] == 1 &&
// CPUID.(EAX = 80000001H, 0H) : ECX.LZCNT[bit 5] == 1 &&
// CPUID.(EAX = 01H, ECX = 0H) : ECX.MOVBE[bit 22] == 1
//
SSELIBRARY_API bool Supports_AVX2()
{
    bool ret = false;

	GetCPUInfo(1, 0);
	bool fmaFlag = g_info.regF1.ecx.Intel.FMA;
	bool movbeFlag = g_info.regF1.ecx.Intel.MOVBE == 1;
    bool osxFlag = g_info.regF1.ecx.Intel.OSXSAVE == 1;

    GetCPUInfo(7,0);
    bool avx2Flag = g_info.regF7.ebx.Intel.AVX2 == 1;
	bool bmi1Flag = g_info.regF7.ebx.Intel.BMI1 == 1;
	bool bmi2Flag = g_info.regF7.ebx.Intel.BMI2 == 1;

	GetCPUInfo(0x80000001, 0);
	bool lzcntFlag = g_info.regF8.ecx.Intel.ABM == 1;	// bit 5

	if ( fmaFlag && movbeFlag && avx2Flag && bmi1Flag && bmi2Flag && lzcntFlag && osxFlag )
	{ 
        // Supported by the CPU, check the OS.
        unsigned long long featureMask = _xgetbv(_XCR_XFEATURE_ENABLED_MASK);
        ret = (featureMask & 0x6) == 0x6;
    }
    return ret;
}

SSELIBRARY_API bool Supports_AVX512()
{
    bool ret = false;
    // TBD: How to check for this.

    return ret;
}

