//////////////////////////////////////////////////////////////////////////////
//   Author:       George Potts
//   Copyright:    PottsSoftware
//   Initial Date: 2015/08/13
// 
//   Description:  Implements many of the operations for 8-bit variables.
//
//   Notes:
//
//   Revision History:
//
//////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "SSELibrary.h"
#include <immintrin.h>

#pragma region Max_U8
SSELIBRARY_API void AVX2_Max_U8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
	__m256i * ptrDst = (__m256i*)dst;
	__m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
	{
		__m256i m_dst = _mm256_load_si256(ptrDst);
		__m256i m_src = _mm256_load_si256(ptrSrc);
		__m256i result = _mm256_max_epu8(m_dst, m_src);
		_mm256_store_si256(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}
SSELIBRARY_API void SSE_Max_U8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    for( ; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);
        __m128i result = _mm_max_epu8(m_dst, m_src);    // SSE2
        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}
SSELIBRARY_API void CPP_Max_U8(char* dst, char* src, size_t bytes)
{
	unsigned char * pdst = (unsigned char*)dst;
	unsigned char * psrc = (unsigned char*)src;
    for (; bytes > 0; bytes--)
	{
        if (*pdst < *psrc) *pdst = *psrc;
        psrc++;
        pdst++;
	}
}
#pragma endregion 

#pragma region Min_U8
SSELIBRARY_API void AVX2_Min_U8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
	__m256i * ptrDst = (__m256i*)dst;
	__m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
	{
		__m256i m_dst = _mm256_load_si256(ptrDst);
		__m256i m_src = _mm256_load_si256(ptrSrc);
		__m256i result = _mm256_min_epu8(m_dst, m_src);
		_mm256_store_si256(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}
SSELIBRARY_API void SSE_Min_U8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    for( ; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);
        __m128i result = _mm_min_epu8(m_dst, m_src);
        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}
SSELIBRARY_API void CPP_Min_U8(char* dst, char* src, size_t bytes)
{
	unsigned char * pdst = (unsigned char*)dst;
	unsigned char * psrc = (unsigned char*)src;
    for (; bytes > 0; bytes--)
	{
        if (*pdst > *psrc) *pdst = *psrc;
        psrc++;
        pdst++;
	}
}
#pragma endregion

#pragma region Max_I8
SSELIBRARY_API void AVX2_Max_I8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
	__m256i * ptrDst = (__m256i*)dst;
	__m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
	{
		__m256i m_dst = _mm256_load_si256(ptrDst);
		__m256i m_src = _mm256_load_si256(ptrSrc);
		__m256i result = _mm256_max_epi8(m_dst, m_src);
		_mm256_store_si256(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}
SSELIBRARY_API void SSE41_Max_I8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    for( ; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);
        __m128i result = _mm_max_epi8(m_dst, m_src);    // SSE41
        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}
SSELIBRARY_API void CPP_Max_I8(char* dst, char* src, size_t bytes)
{
	char * pdst = (char*)dst;
	char * psrc = (char*)src;
    for (; bytes > 0; bytes--)
	{
        if (*pdst < *psrc) *pdst = *psrc;
        psrc++;
        pdst++;
	}
}
#pragma endregion

#pragma region Min_I8
SSELIBRARY_API void AVX2_Min_I8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
	__m256i * ptrDst = (__m256i*)dst;
	__m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
	{
		__m256i m_dst = _mm256_load_si256(ptrDst);
		__m256i m_src = _mm256_load_si256(ptrSrc);
		__m256i result = _mm256_min_epi8(m_dst, m_src);
		_mm256_store_si256(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}
SSELIBRARY_API void SSE41_Min_I8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    for (; loops > 0;  loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);
        __m128i result = _mm_min_epi8(m_dst, m_src);
        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}
SSELIBRARY_API void CPP_Min_I8(char* dst, char* src, size_t bytes)
{
	char * pdst = (char*)dst;
	char * psrc = (char*)src;
    for (; bytes > 0; bytes--)
	{
        if (*pdst > *psrc) *pdst = *psrc;
        psrc++;
        pdst++;
	}
}
#pragma endregion

#pragma region Diff_U8
SSELIBRARY_API void AVX2_Diff_U8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
	__m256i * ptrDst = (__m256i*)dst;
	__m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
	{
		__m256i m_dst = _mm256_load_si256(ptrDst);
		__m256i m_src = _mm256_load_si256(ptrSrc);
        __m256i MAX = _mm256_max_epu8(m_dst, m_src);
        __m256i MIN = _mm256_min_epu8(m_dst, m_src);
        __m256i result = _mm256_subs_epu8(MAX, MIN);
		_mm256_store_si256(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}
SSELIBRARY_API void SSE_Diff_U8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    for( ; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);
        __m128i MAX = _mm_max_epu8(m_dst, m_src);
        __m128i MIN = _mm_min_epu8(m_dst, m_src);
        __m128i result = _mm_subs_epu8(MAX,MIN);
        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}
SSELIBRARY_API void CPP_Diff_U8(char* dst, char* src, size_t bytes)
{
	unsigned char * pdst = (unsigned char*)dst;
	unsigned char * psrc = (unsigned char*)src;
    for (; bytes > 0; bytes--)
	{
        if (*pdst < *psrc) *pdst = (unsigned char)(*psrc - *pdst);
        else *pdst = (unsigned char)(*pdst - *psrc);
        psrc++;
        pdst++;
	}
}
#pragma endregion

#pragma region Diff_I8
SSELIBRARY_API void AVX2_Diff_I8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
	__m256i * ptrDst = (__m256i*)dst;
	__m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
	{
		__m256i m_dst = _mm256_load_si256(ptrDst);
		__m256i m_src = _mm256_load_si256(ptrSrc);
        __m256i MAX = _mm256_max_epi8(m_dst, m_src);
        __m256i MIN = _mm256_min_epi8(m_dst, m_src);
        __m256i result = _mm256_subs_epi8(MAX, MIN);
		_mm256_store_si256(ptrDst, result);
		ptrDst++;
		ptrSrc++;
	}
}

SSELIBRARY_API void SSE_Diff_I8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    for( ; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);
        __m128i MAX = _mm_max_epi8(m_dst, m_src);
        __m128i MIN = _mm_min_epi8(m_dst, m_src);
        __m128i result = _mm_sub_epi8(MAX,MIN);
        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}
SSELIBRARY_API void CPP_Diff_I8(char* dst, char* src, size_t bytes)
{
	signed char * pdst = (signed char*)dst;
	signed char * psrc = (signed char*)src;
    for (; bytes > 0; bytes--)
	{
        if (*pdst < *psrc) *pdst = *psrc - *pdst;
        else *pdst = *pdst - *psrc;
        psrc++;
        pdst++;
	}
}
#pragma endregion